#include "Instance.h"

CInstance::CInstance(FILE *file)
{	
	fscanf_s(file, "%d %d %d", &id, &n, &M);
	this->items.LoadItems(file, n);
	best = new bool[n];
	Clear(best, n);
	theBest = 0;
}

void CInstance::RecBB(int pos, int Mleft, int Cin, int Cleft, bool *actual)
{
	if(Mleft<0) // Too manu items in the bag (Branch)
		return;

	if((Cin+Cleft)<theBest) // Can't make it better
		return;

	if(Cin>theBest) // Current is better than the best
	{
		theBest = Cin;
		for(int i=0; i<n; i++)
			best[i]=actual[i];
	}
	if(pos<0) // Finished the items
		return;

	//Include
	actual[pos]=true;
	RecBB(pos-1, Mleft-items.GetItem(pos).GetWeight(), Cin+items.GetItem(pos).GetCost(),
		Cleft-items.GetItem(pos).GetCost(), actual);
	//Not include
	actual[pos]=false;
	RecBB(pos-1, Mleft, Cin,
		Cleft-items.GetItem(pos).GetCost(), actual);

}

float CInstance::SolveBB()
{
	float t1, t2;
	int rep;
	bool *actual = new bool[n];	

	rep = 0;
	t1 = GetMyCPUTime();

	do{
		Clear(best, n);
		Clear(actual, n);
		theBest = 0;
		RecBB(n-1, M, 0, items.GetTotalCost(), actual);
		rep++;
		t2 = GetMyCPUTime();
	}while(t2-t1 < 100);

	delete [] actual;
	return (t2-t1)/rep;
}

void CInstance::Heur2()
{
	priority_queue<CItem> sortedItems;
	int Mleft;
	CItem current;
	CItem bestFit;

	for(int i=0; i<items.GetNoItems(); i++)
	{
		sortedItems.push(items.GetItem(i));
	}

	Mleft = M;

	bestFit = sortedItems.top();
	while(sortedItems.size() != 0){
		current = sortedItems.top();
		sortedItems.pop();
		if(current.GetWeight() <= Mleft){
			Mleft -= current.GetWeight();
			theBest += current.GetCost();
			best[current.GetPos()] = true;
		}
		if( current.GetWeight() <= M){
			if( current.GetCost() > bestFit.GetCost() )
				bestFit = current;
		}
	}

	if(bestFit.GetCost() > theBest){
		Clear(best, n);
		best[bestFit.GetPos()] = true;
		theBest = bestFit.GetCost();
	}

}

float CInstance::SolveHeur2()
{
	float t1, t2;
	int rep;

	rep = 0;
	t1 = GetMyCPUTime();

	do{
		Clear(best, n);
		theBest = 0;
		Heur2();
		rep++;
		t2 = GetMyCPUTime();
	}while(t2-t1 < 100);

	return (t2-t1)/rep;
}

void CInstance::Dyn()
{
	// use map(col, row) to access elements;
	// columns are items, rows are weights;
	int *totCost = new int[n*(M+1)];
	int *totWeight = new int[n*(M+1)];
	bool *isIn = new bool[n*(M+1)];
	int costIn;

	//Set column 0 to the first item;
	for(int j=0; j<(M+1); j++)
	{
		if(j>=items.GetItem(0).GetWeight()){ //the item 0 fits in
			totCost[map(0, j)] = items.GetItem(0).GetCost();
			totWeight[map(0, j)] = items.GetItem(0).GetWeight();
			isIn[map(0, j)] = true;
		}
		else{
			totCost[map(0, j)] = 0;
			totWeight[map(0, j)] = 0;
			isIn[map(0, j)] = false;
		}
	}

	for(int i=1; i<n; i++){
		for(int j=0; j<(M+1); j++){
			if(j-items.GetItem(i).GetWeight() < 0){ 
				//ie. this items won't fit in and we copy the best for this M
				totCost[map(i, j)] = totCost[map(i-1, j)];
				totWeight[map(i, j)] = totWeight[map(i-1, j)];
				isIn[map(i, j)] = false;
			}
			else{ // have to compare the best solution
				// the best cost available if the item is in;
				costIn = totCost[map(i-1, j-items.GetItem(i).GetWeight())] + items.GetItem(i).GetCost();
				if(totCost[map(i-1, j)] > costIn){ //putting it in ain't in out advantage
					totCost[map(i, j)] = totCost[map(i-1, j)];
					totWeight[map(i, j)] = totWeight[map(i-1, j)];
					isIn[map(i, j)] = false;
				}else{
					totCost[map(i, j)] = costIn;
					totWeight[map(i, j)] = totWeight[map(i-1, j-items.GetItem(i).GetWeight())];
					isIn[map(i, j)] = true;
				}
			}
		}
	}
	//backtracking
	int j=M;
	for(int i=n-1; i>=0; i--){
		best[i] = isIn[map(i,j)];
		if(best[i]){
			j-=items.GetItem(i).GetWeight();
			theBest+=items.GetItem(i).GetCost();
		}
	}

	delete [] totCost;
	delete [] totWeight;
	delete [] isIn;
}

float CInstance::SolveDyn()
{
	float t1, t2;
	int rep;


	rep = 0;
	t1 = GetMyCPUTime();

	do{
		Clear(best, n);
		theBest = 0;
		Dyn();
		rep++;
		t2 = GetMyCPUTime();
	}while(t2-t1 < 100);

	return (t2-t1)/rep;
}
void CInstance::Clear(bool *sol, int n)
{
	for(int i=0; i<n; i++)
		sol[i] = false;
}


int CInstance::GetCapacity()
{
	return this->M;
}

int CInstance::GetID()
{
	return this->id;
}

int CInstance::GetNoElements()
{
	return this->n;
}

bool *CInstance::GetCombSolution()
{
	return this->best;
}

int CInstance::GetBestCost()
{
	return this->theBest;
}


float CInstance::GetMyCPUTime()
{
    HANDLE hProcess;    // specifies the process of interest
    FILETIME lpCreationTime;    // when the process was created
    FILETIME lpExitTime;    // when the process exited
    FILETIME lpKernelTime;    // time the process has spent in kernel mode
    FILETIME lpUserTime;
    LARGE_INTEGER t1, t2;


    hProcess = GetCurrentProcess();
    GetProcessTimes(hProcess, &lpCreationTime, &lpExitTime, &lpKernelTime, &lpUserTime);
    memcpy(&t1, &lpUserTime, sizeof(FILETIME));
    memcpy(&t2, &lpKernelTime, sizeof(FILETIME));
    return ((float)(t1.QuadPart+t2.QuadPart)/10000);
}


int CInstance::map(const int &col, const int &row)
{
	return row*n+col;
}


CInstance::~CInstance(void)
{
	delete [] best;
}

ostream& operator<<(ostream& x, const CInstance &a)
{
	x << a.id << "\t" << a.n << "\t" << a.theBest;
	for(int i=0; i<a.n; i++)
	{
		if(a.best[i])
			x << "\t1";
		else
			x << "\t0";
	}
	return x;
}